@@ -83,13 +83,25 @@ class FrameworkLoopTransformer(FrameworkTransformer):
                 )
             stage: Transformer
             for stage in stages:
+                if hasattr(stage, "getName"):
+                    # noinspection Mypy
+                    stage_name = stage.getName()
+                else:
+                    stage_name = stage.__class__.__name__
                 if progress_logger is not None:
                     progress_logger.start_mlflow_run(
-                        run_name=str(stage), is_nested=True
+                        run_name=stage_name, is_nested=True
                     )
                 if hasattr(stage, "set_loop_id"):
                     stage.set_loop_id(str(current_run_number))
-                df = stage.transform(df)
+
+                try:
+                    df = stage.transform(df)
+                except Exception as e:
+                    if len(e.args) >= 1:
+                        # e.args = (e.args[0] + f" in stage {stage_name}") + e.args[1:]
+                        e.args = (f"In Stage ({stage_name})", *e.args)
+                    raise e
                 if progress_logger is not None:
                     progress_logger.end_mlflow_run()
             time_elapsed: float = time.time() - start_time
