@@ -8,6 +8,7 @@ from spark_pipeline_framework.utilities.dictionary_writer.v1.dictionary_writer i
 class FhirReceiverException(Exception):
     def __init__(
         self,
+        *,
         url: str,
         json_data: str,
         response_text: Optional[str],
@@ -17,7 +18,7 @@ class FhirReceiverException(Exception):
         self.url: str = url
         self.data: str = json_data
         json = {
-            "message": f"FHIR send failed: {message}",
+            "message": f"FHIR receive failed: {message}",
             "url": url,
             "status_code": response_status_code,
             "response_text": response_text,
