@@ -1,12 +1,22 @@
+import asyncio
 import json
 from datetime import datetime
 from json import JSONDecodeError
 from logging import Logger
-from typing import Any, Dict, Iterable, List, Optional, Union, cast, NamedTuple
+from typing import (
+    Any,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Union,
+    cast,
+    NamedTuple,
+    Coroutine,
+)
 
 from furl import furl
 
-# noinspection PyPep8Naming
 from helix_fhir_client_sdk.exceptions.fhir_sender_exception import FhirSenderException
 from helix_fhir_client_sdk.fhir_client import FhirClient
 from helix_fhir_client_sdk.filters.sort_field import SortField
@@ -25,6 +35,7 @@ from spark_pipeline_framework.utilities.fhir_helpers.fhir_receiver_exception imp
 from spark_pipeline_framework.utilities.fhir_helpers.get_fhir_client import (
     get_fhir_client,
 )
+from spark_pipeline_framework.utilities.flattener.flattener import flatten
 
 
 class GetBatchResult(NamedTuple):
@@ -68,7 +79,8 @@ class FhirReceiverHelpers:
         resource_type: str,
         error_view: Optional[str],
         url_column: Optional[str],
-    ) -> Iterable[Row]:
+        use_data_streaming: Optional[bool],
+    ) -> List[Row]:
         resource_id_with_token_list: List[Dict[str, Optional[str]]] = [
             {
                 "resource_id": r["id"],
@@ -86,7 +98,7 @@ class FhirReceiverHelpers:
             else {"resource_id": r["id"], "access_token": auth_access_token}
             for r in rows
         ]
-        yield from FhirReceiverHelpers.process_with_token(
+        result = FhirReceiverHelpers.process_with_token(
             partition_index=partition_index,
             resource_id_with_token_list=resource_id_with_token_list,
             batch_size=batch_size,
@@ -118,7 +130,9 @@ class FhirReceiverHelpers:
             resource_type=resource_type,
             error_view=error_view,
             url_column=url_column,
+            use_data_streaming=use_data_streaming,
         )
+        return result
 
     @staticmethod
     def process_with_token(
@@ -154,7 +168,8 @@ class FhirReceiverHelpers:
         resource_type: str,
         error_view: Optional[str],
         url_column: Optional[str],
-    ) -> Iterable[Row]:
+        use_data_streaming: Optional[bool],
+    ) -> List[Row]:
         try:
             first_id: Optional[str] = resource_id_with_token_list[0]["resource_id"]
         except IndexError:
@@ -168,22 +183,23 @@ class FhirReceiverHelpers:
         sent: int = len(resource_id_with_token_list)
 
         if sent == 0:
-            yield Row(
-                partition_index=partition_index,
-                sent=0,
-                received=0,
-                responses=[],
-                first=None,
-                last=None,
-                error_text=None,
-                url=None,
-                status_code=None,
-            )
-            return
+            return [
+                Row(
+                    partition_index=partition_index,
+                    sent=0,
+                    received=0,
+                    responses=[],
+                    first=None,
+                    last=None,
+                    error_text=None,
+                    url=None,
+                    status_code=None,
+                )
+            ]
 
         # if batch and not has_token then send all ids at once as long as the access token is the same
         if batch_size and batch_size > 1 and not has_token_col:
-            yield from FhirReceiverHelpers.process_batch(
+            return FhirReceiverHelpers.process_batch(
                 partition_index=partition_index,
                 first_id=first_id,
                 last_id=last_id,
@@ -214,9 +230,10 @@ class FhirReceiverHelpers:
                 auth_access_token=auth_access_token,
                 resource_type=resource_type,
                 error_view=error_view,
+                use_data_streaming=use_data_streaming,
             )
         else:  # otherwise send one by one
-            yield from FhirReceiverHelpers.process_one_by_one(
+            return FhirReceiverHelpers.process_one_by_one(
                 partition_index=partition_index,
                 first_id=first_id,
                 last_id=last_id,
@@ -247,6 +264,7 @@ class FhirReceiverHelpers:
                 error_view=error_view,
                 url_column=url_column,
                 resource_name=resource_type,
+                use_data_streaming=use_data_streaming,
             )
 
     @staticmethod
@@ -282,69 +300,150 @@ class FhirReceiverHelpers:
         error_view: Optional[str],
         url_column: Optional[str],
         resource_name: str,
-    ) -> Iterable[Row]:
-        for resource1 in resource_id_with_token_list:
-            id_ = resource1["resource_id"]
-            token_ = resource1["access_token"]
-            url_ = resource1.get(url_column) if url_column else None
-            service_slug = resource1.get(slug_column) if slug_column else None
-            resource_type = resource1.get("resourceType")
-            responses_from_fhir: List[str] = []
-            try:
-                result1 = FhirReceiverHelpers.send_simple_fhir_request(
-                    id_=id_,
-                    token_=token_,
-                    server_url_=url_ or server_url,
-                    service_slug=service_slug,
-                    resource_type=resource_type or resource_name,
-                    log_level=log_level,
-                    server_url=server_url,
-                    action=action,
-                    action_payload=action_payload,
-                    additional_parameters=additional_parameters,
-                    filter_by_resource=filter_by_resource,
-                    filter_parameter=filter_parameter,
-                    sort_fields=sort_fields,
-                    auth_server_url=auth_server_url,
-                    auth_client_id=auth_client_id,
-                    auth_client_secret=auth_client_secret,
-                    auth_login_token=auth_login_token,
-                    auth_scopes=auth_scopes,
-                    include_only_properties=include_only_properties,
-                    separate_bundle_resources=separate_bundle_resources,
-                    expand_fhir_bundle=expand_fhir_bundle,
-                    accept_type=accept_type,
-                    content_type=content_type,
-                    accept_encoding=accept_encoding,
-                    slug_column=slug_column,
-                    retry_count=retry_count,
-                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
-                    limit=limit,
-                )
-                resp_result: str = result1.responses.replace("\n", "")
-                try:
-                    responses_from_fhir = FhirReceiverHelpers.json_str_to_list_str(
-                        resp_result
-                    )
-                except JSONDecodeError as e2:
-                    if error_view:
-                        result1.error = f"{(result1.error or '')}: {str(e2)}"
-                    else:
-                        raise FhirParserException(
-                            url=result1.url,
-                            message="Parsing result as json failed",
-                            json_data=result1.responses,
-                            response_status_code=result1.status,
-                        ) from e2
+        use_data_streaming: Optional[bool],
+    ) -> List[Row]:
+        coroutines: List[Coroutine[Any, Any, List[Row]]] = [
+            FhirReceiverHelpers.process_single_row_async(
+                partition_index=partition_index,
+                first_id=first_id,
+                last_id=last_id,
+                server_url=server_url,
+                log_level=log_level,
+                action=action,
+                action_payload=action_payload,
+                additional_parameters=additional_parameters,
+                filter_by_resource=filter_by_resource,
+                filter_parameter=filter_parameter,
+                sort_fields=sort_fields,
+                auth_server_url=auth_server_url,
+                auth_client_id=auth_client_id,
+                auth_client_secret=auth_client_secret,
+                auth_login_token=auth_login_token,
+                auth_scopes=auth_scopes,
+                include_only_properties=include_only_properties,
+                separate_bundle_resources=separate_bundle_resources,
+                expand_fhir_bundle=expand_fhir_bundle,
+                accept_type=accept_type,
+                content_type=content_type,
+                accept_encoding=accept_encoding,
+                slug_column=slug_column,
+                retry_count=retry_count,
+                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
+                limit=limit,
+                error_view=error_view,
+                url_column=url_column,
+                resource1=resource1,
+                resource_name=resource_name,
+                use_data_streaming=use_data_streaming,
+            )
+            for resource1 in resource_id_with_token_list
+        ]
 
-                error_text = result1.error
-                status_code = result1.status
-                request_url = result1.url
-            except FhirSenderException as e1:
-                error_text = str(e1)
-                status_code = e1.response_status_code or 0
-                request_url = e1.url
-            yield Row(
+        async def get_functions() -> List[List[Row]]:
+            # noinspection PyTypeChecker
+            return await asyncio.gather(*[asyncio.create_task(c) for c in coroutines])
+
+        result: List[List[Row]] = asyncio.run(get_functions())
+        return flatten(result)
+
+    @staticmethod
+    async def process_single_row_async(
+        *,
+        partition_index: int,
+        first_id: Optional[str],
+        last_id: Optional[str],
+        server_url: Optional[str],
+        log_level: Optional[str],
+        action: Optional[str],
+        action_payload: Optional[Dict[str, Any]],
+        additional_parameters: Optional[List[str]],
+        filter_by_resource: Optional[str],
+        filter_parameter: Optional[str],
+        sort_fields: Optional[List[SortField]],
+        auth_server_url: Optional[str],
+        auth_client_id: Optional[str],
+        auth_client_secret: Optional[str],
+        auth_login_token: Optional[str],
+        auth_scopes: Optional[List[str]],
+        include_only_properties: Optional[List[str]],
+        separate_bundle_resources: bool,
+        expand_fhir_bundle: bool,
+        accept_type: Optional[str],
+        content_type: Optional[str],
+        accept_encoding: Optional[str],
+        slug_column: Optional[str],
+        retry_count: Optional[int],
+        exclude_status_codes_from_retry: Optional[List[int]],
+        limit: Optional[int],
+        error_view: Optional[str],
+        url_column: Optional[str],
+        resource_name: str,
+        resource1: Dict[str, Optional[str]],
+        use_data_streaming: Optional[bool],
+    ) -> List[Row]:
+        id_ = resource1["resource_id"]
+        token_ = resource1["access_token"]
+        url_ = resource1.get(url_column) if url_column else None
+        service_slug = resource1.get(slug_column) if slug_column else None
+        resource_type = resource1.get("resourceType")
+        responses_from_fhir: List[str] = []
+        try:
+            result1 = await FhirReceiverHelpers.send_simple_fhir_request_async(
+                id_=id_,
+                token_=token_,
+                server_url_=url_ or server_url,
+                service_slug=service_slug,
+                resource_type=resource_type or resource_name,
+                log_level=log_level,
+                server_url=server_url,
+                action=action,
+                action_payload=action_payload,
+                additional_parameters=additional_parameters,
+                filter_by_resource=filter_by_resource,
+                filter_parameter=filter_parameter,
+                sort_fields=sort_fields,
+                auth_server_url=auth_server_url,
+                auth_client_id=auth_client_id,
+                auth_client_secret=auth_client_secret,
+                auth_login_token=auth_login_token,
+                auth_scopes=auth_scopes,
+                include_only_properties=include_only_properties,
+                separate_bundle_resources=separate_bundle_resources,
+                expand_fhir_bundle=expand_fhir_bundle,
+                accept_type=accept_type,
+                content_type=content_type,
+                accept_encoding=accept_encoding,
+                slug_column=slug_column,
+                retry_count=retry_count,
+                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
+                limit=limit,
+                use_data_streaming=use_data_streaming,
+            )
+            resp_result: str = result1.responses.replace("\n", "")
+            try:
+                responses_from_fhir = FhirReceiverHelpers.json_str_to_list_str(
+                    resp_result
+                )
+            except JSONDecodeError as e2:
+                if error_view:
+                    result1.error = f"{(result1.error or '')}: {str(e2)}"
+                else:
+                    raise FhirParserException(
+                        url=result1.url,
+                        message="Parsing result as json failed",
+                        json_data=result1.responses,
+                        response_status_code=result1.status,
+                    ) from e2
+
+            error_text = result1.error
+            status_code = result1.status
+            request_url = result1.url
+        except FhirSenderException as e1:
+            error_text = str(e1)
+            status_code = e1.response_status_code or 0
+            request_url = e1.url
+        result = [
+            Row(
                 partition_index=partition_index,
                 sent=1,
                 received=len(responses_from_fhir),
@@ -355,6 +454,8 @@ class FhirReceiverHelpers:
                 url=request_url,
                 status_code=status_code,
             )
+        ]
+        return result
 
     @staticmethod
     def process_batch(
@@ -389,35 +490,39 @@ class FhirReceiverHelpers:
         auth_access_token: Optional[str],
         resource_type: str,
         error_view: Optional[str],
-    ) -> Iterable[Row]:
-        result1 = FhirReceiverHelpers.send_simple_fhir_request(
-            id_=[cast(str, r["resource_id"]) for r in resource_id_with_token_list],
-            token_=auth_access_token,
-            server_url=server_url,
-            server_url_=server_url,
-            resource_type=resource_type,
-            log_level=log_level,
-            action=action,
-            action_payload=action_payload,
-            additional_parameters=additional_parameters,
-            filter_by_resource=filter_by_resource,
-            filter_parameter=filter_parameter,
-            sort_fields=sort_fields,
-            auth_server_url=auth_server_url,
-            auth_client_id=auth_client_id,
-            auth_client_secret=auth_client_secret,
-            auth_login_token=auth_login_token,
-            auth_scopes=auth_scopes,
-            include_only_properties=include_only_properties,
-            separate_bundle_resources=separate_bundle_resources,
-            expand_fhir_bundle=expand_fhir_bundle,
-            accept_type=accept_type,
-            content_type=content_type,
-            accept_encoding=accept_encoding,
-            slug_column=slug_column,
-            retry_count=retry_count,
-            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
-            limit=limit,
+        use_data_streaming: Optional[bool],
+    ) -> List[Row]:
+        result1 = asyncio.run(
+            FhirReceiverHelpers.send_simple_fhir_request_async(
+                id_=[cast(str, r["resource_id"]) for r in resource_id_with_token_list],
+                token_=auth_access_token,
+                server_url=server_url,
+                server_url_=server_url,
+                resource_type=resource_type,
+                log_level=log_level,
+                action=action,
+                action_payload=action_payload,
+                additional_parameters=additional_parameters,
+                filter_by_resource=filter_by_resource,
+                filter_parameter=filter_parameter,
+                sort_fields=sort_fields,
+                auth_server_url=auth_server_url,
+                auth_client_id=auth_client_id,
+                auth_client_secret=auth_client_secret,
+                auth_login_token=auth_login_token,
+                auth_scopes=auth_scopes,
+                include_only_properties=include_only_properties,
+                separate_bundle_resources=separate_bundle_resources,
+                expand_fhir_bundle=expand_fhir_bundle,
+                accept_type=accept_type,
+                content_type=content_type,
+                accept_encoding=accept_encoding,
+                slug_column=slug_column,
+                retry_count=retry_count,
+                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
+                limit=limit,
+                use_data_streaming=use_data_streaming,
+            )
         )
         resp_result: str = result1.responses.replace("\n", "")
         responses_from_fhir = []
@@ -437,20 +542,23 @@ class FhirReceiverHelpers:
         error_text = result1.error
         status_code = result1.status
         is_valid_response: bool = True if len(responses_from_fhir) > 0 else False
-        yield Row(
-            partition_index=partition_index,
-            sent=1,
-            received=len(responses_from_fhir) if is_valid_response else 0,
-            responses=responses_from_fhir if is_valid_response else [],
-            first=first_id,
-            last=last_id,
-            error_text=error_text,
-            url=result1.url,
-            status_code=status_code,
-        )
+        result = [
+            Row(
+                partition_index=partition_index,
+                sent=1,
+                received=len(responses_from_fhir) if is_valid_response else 0,
+                responses=responses_from_fhir if is_valid_response else [],
+                first=first_id,
+                last=last_id,
+                error_text=error_text,
+                url=result1.url,
+                status_code=status_code,
+            )
+        ]
+        return result
 
     @staticmethod
-    def send_simple_fhir_request(
+    async def send_simple_fhir_request_async(
         *,
         id_: Optional[Union[str, List[str]]],
         token_: Optional[str],
@@ -480,10 +588,11 @@ class FhirReceiverHelpers:
         retry_count: Optional[int],
         exclude_status_codes_from_retry: Optional[List[int]],
         limit: Optional[int],
+        use_data_streaming: Optional[bool],
     ) -> FhirGetResponse:
         url = server_url_ or server_url
         assert url
-        return FhirReceiverHelpers.send_fhir_request(
+        return await FhirReceiverHelpers.send_fhir_request_async(
             logger=get_logger(__name__),
             action=action,
             action_payload=action_payload,
@@ -513,10 +622,12 @@ class FhirReceiverHelpers:
             exclude_status_codes_from_retry=exclude_status_codes_from_retry,
             limit=limit,
             log_level=log_level,
+            use_data_streaming=use_data_streaming,
         )
 
     @staticmethod
-    def send_fhir_request(
+    async def send_fhir_request_async(
+        *,
         logger: Logger,
         action: Optional[str],
         action_payload: Optional[Dict[str, Any]],
@@ -548,6 +659,7 @@ class FhirReceiverHelpers:
         retry_count: Optional[int] = None,
         exclude_status_codes_from_retry: Optional[List[int]] = None,
         limit: Optional[int] = None,
+        use_data_streaming: Optional[bool],
     ) -> FhirGetResponse:
         """
         Sends a fhir request to the fhir client sdk
@@ -584,6 +696,7 @@ class FhirReceiverHelpers:
         :param exclude_status_codes_from_retry: do not retry for these status codes
         :param limit: limit results to this
         :param log_level:
+        :param use_data_streaming:
         :return:
         :rtype:
         """
@@ -652,7 +765,11 @@ class FhirReceiverHelpers:
 
         if limit is not None:
             fhir_client = fhir_client.limit(limit=limit)
-        return fhir_client.get()
+
+        if use_data_streaming:
+            fhir_client = fhir_client.use_data_streaming(use_data_streaming)
+
+        return await fhir_client.get_async()
 
     @staticmethod
     def json_str_to_list_str(json_str: str) -> List[str]:
@@ -699,6 +816,7 @@ class FhirReceiverHelpers:
         log_level: Optional[str],
         error_view: Optional[str],
         ignore_status_codes: List[int],
+        use_data_streaming: Optional[bool],
     ) -> GetBatchResult:
         resources: List[str] = []
         errors: List[str] = []
@@ -708,43 +826,41 @@ class FhirReceiverHelpers:
         page_number: int = 0
         server_page_number: int = 0
         assert server_url
-        while True:
-            result = FhirReceiverHelpers.send_fhir_request(
-                logger=get_logger(__name__),
-                action=action,
-                action_payload=action_payload,
-                additional_parameters=additional_parameters,
-                filter_by_resource=filter_by_resource,
-                filter_parameter=filter_parameter,
-                resource_name=resource_name,
-                resource_id=None,
-                server_url=server_url,
-                include_only_properties=include_only_properties,
-                page_number=server_page_number,  # since we're setting id:above we can leave this as 0
-                page_size=page_size,
-                last_updated_after=last_updated_after,
-                last_updated_before=last_updated_before,
-                sort_fields=sort_fields,
-                auth_server_url=auth_server_url,
-                auth_client_id=auth_client_id,
-                auth_client_secret=auth_client_secret,
-                auth_login_token=auth_login_token,
-                auth_access_token=auth_access_token,
-                auth_scopes=auth_scopes,
-                separate_bundle_resources=separate_bundle_resources,
-                expand_fhir_bundle=expand_fhir_bundle,
-                accept_type=accept_type,
-                content_type=content_type,
-                accept_encoding=accept_encoding,
-                retry_count=retry_count,
-                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
-                log_level=log_level,
-            )
-            result_response: List[str] = []
-            try:
-                result_response = FhirReceiverHelpers.json_str_to_list_str(
-                    result.responses
+        if use_data_streaming:
+            result = asyncio.run(
+                FhirReceiverHelpers.send_fhir_request_async(
+                    logger=get_logger(__name__),
+                    action=action,
+                    action_payload=action_payload,
+                    additional_parameters=additional_parameters,
+                    filter_by_resource=filter_by_resource,
+                    filter_parameter=filter_parameter,
+                    resource_name=resource_name,
+                    resource_id=None,
+                    server_url=server_url,
+                    include_only_properties=include_only_properties,
+                    last_updated_after=last_updated_after,
+                    last_updated_before=last_updated_before,
+                    sort_fields=sort_fields,
+                    auth_server_url=auth_server_url,
+                    auth_client_id=auth_client_id,
+                    auth_client_secret=auth_client_secret,
+                    auth_login_token=auth_login_token,
+                    auth_access_token=auth_access_token,
+                    auth_scopes=auth_scopes,
+                    separate_bundle_resources=separate_bundle_resources,
+                    expand_fhir_bundle=expand_fhir_bundle,
+                    accept_type=accept_type,
+                    content_type=content_type,
+                    accept_encoding=accept_encoding,
+                    retry_count=retry_count,
+                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
+                    log_level=log_level,
+                    use_data_streaming=use_data_streaming,
                 )
+            )
+            try:
+                resources = FhirReceiverHelpers.json_str_to_list_str(result.responses)
             except JSONDecodeError as e:
                 if error_view:
                     errors.append(
@@ -764,62 +880,122 @@ class FhirReceiverHelpers:
                         json_data=result.responses,
                         response_status_code=result.status,
                     ) from e
-
-            auth_access_token = result.access_token
-            if len(result_response) > 0:
-                # get id of last resource
-                json_resources: List[Dict[str, Any]] = json.loads(result.responses)
-                if isinstance(json_resources, list):  # normal response
-                    if len(json_resources) > 0:  # received any resources back
-                        last_json_resource = json_resources[-1]
-                        if result.next_url:
-                            # if server has sent back a next url then use that
-                            next_url: Optional[str] = result.next_url
-                            next_uri: furl = furl(next_url)
-                            additional_parameters = [
-                                f"{k}={v}" for k, v in next_uri.args.items()
-                            ]
-                            # remove any entry for id:above
-                            additional_parameters = list(
-                                filter(
-                                    lambda x: not x.startswith("_count")
-                                    and not x.startswith("_element"),
-                                    additional_parameters,
-                                )
+        else:
+            while True:
+                result = asyncio.run(
+                    FhirReceiverHelpers.send_fhir_request_async(
+                        logger=get_logger(__name__),
+                        action=action,
+                        action_payload=action_payload,
+                        additional_parameters=additional_parameters,
+                        filter_by_resource=filter_by_resource,
+                        filter_parameter=filter_parameter,
+                        resource_name=resource_name,
+                        resource_id=None,
+                        server_url=server_url,
+                        include_only_properties=include_only_properties,
+                        page_number=server_page_number,  # since we're setting id:above we can leave this as 0
+                        page_size=page_size,
+                        last_updated_after=last_updated_after,
+                        last_updated_before=last_updated_before,
+                        sort_fields=sort_fields,
+                        auth_server_url=auth_server_url,
+                        auth_client_id=auth_client_id,
+                        auth_client_secret=auth_client_secret,
+                        auth_login_token=auth_login_token,
+                        auth_access_token=auth_access_token,
+                        auth_scopes=auth_scopes,
+                        separate_bundle_resources=separate_bundle_resources,
+                        expand_fhir_bundle=expand_fhir_bundle,
+                        accept_type=accept_type,
+                        content_type=content_type,
+                        accept_encoding=accept_encoding,
+                        retry_count=retry_count,
+                        exclude_status_codes_from_retry=exclude_status_codes_from_retry,
+                        log_level=log_level,
+                        use_data_streaming=use_data_streaming,
+                    )
+                )
+                result_response: List[str] = []
+                try:
+                    result_response = FhirReceiverHelpers.json_str_to_list_str(
+                        result.responses
+                    )
+                except JSONDecodeError as e:
+                    if error_view:
+                        errors.append(
+                            json.dumps(
+                                {
+                                    "url": result.url,
+                                    "status_code": result.status,
+                                    "error_text": str(e) + " : " + result.responses,
+                                },
+                                default=str,
                             )
-                        elif "id" in last_json_resource:
-                            # use id:above to optimize the next query
-                            id_of_last_resource = last_json_resource["id"]
-                            if not additional_parameters:
-                                additional_parameters = []
-                            # remove any entry for id:above
-                            additional_parameters = list(
-                                filter(
-                                    lambda x: not x.startswith("id:above"),
-                                    additional_parameters,
-                                )
-                            )
-                            additional_parameters.append(
-                                f"id:above={id_of_last_resource}"
-                            )
-                        else:
-                            server_page_number += 1
-                        resources = resources + result_response
-                    page_number += 1
-                    if limit and limit > 0:
-                        if not page_size or (page_number * page_size) >= limit:
-                            break
-                else:
-                    # Received an error
-                    if not result.status in ignore_status_codes:
-                        raise FhirReceiverException(
-                            url=result.url,
-                            json_data=result.responses,
-                            response_text=result.responses,
-                            response_status_code=result.status,
-                            message="Error from FHIR server",
                         )
-            else:
-                break
+                    else:
+                        raise FhirParserException(
+                            url=result.url,
+                            message="Parsing result as json failed",
+                            json_data=result.responses,
+                            response_status_code=result.status,
+                        ) from e
+
+                auth_access_token = result.access_token
+                if len(result_response) > 0:
+                    # get id of last resource
+                    json_resources: List[Dict[str, Any]] = json.loads(result.responses)
+                    if isinstance(json_resources, list):  # normal response
+                        if len(json_resources) > 0:  # received any resources back
+                            last_json_resource = json_resources[-1]
+                            if result.next_url:
+                                # if server has sent back a next url then use that
+                                next_url: Optional[str] = result.next_url
+                                next_uri: furl = furl(next_url)
+                                additional_parameters = [
+                                    f"{k}={v}" for k, v in next_uri.args.items()
+                                ]
+                                # remove any entry for id:above
+                                additional_parameters = list(
+                                    filter(
+                                        lambda x: not x.startswith("_count")
+                                        and not x.startswith("_element"),
+                                        additional_parameters,
+                                    )
+                                )
+                            elif "id" in last_json_resource:
+                                # use id:above to optimize the next query
+                                id_of_last_resource = last_json_resource["id"]
+                                if not additional_parameters:
+                                    additional_parameters = []
+                                # remove any entry for id:above
+                                additional_parameters = list(
+                                    filter(
+                                        lambda x: not x.startswith("id:above"),
+                                        additional_parameters,
+                                    )
+                                )
+                                additional_parameters.append(
+                                    f"id:above={id_of_last_resource}"
+                                )
+                            else:
+                                server_page_number += 1
+                            resources = resources + result_response
+                        page_number += 1
+                        if limit and limit > 0:
+                            if not page_size or (page_number * page_size) >= limit:
+                                break
+                    else:
+                        # Received an error
+                        if not result.status in ignore_status_codes:
+                            raise FhirReceiverException(
+                                url=result.url,
+                                json_data=result.responses,
+                                response_text=result.responses,
+                                response_status_code=result.status,
+                                message="Error from FHIR server",
+                            )
+                else:
+                    break
 
         return GetBatchResult(resources=resources, errors=errors)
